/*
 * Программа дистанционного управления 
 * бытовым вентилятором по WiFi
 * на основе NodeMCU v3 / ESP8266
 * Панкратьев Д.И., 2020
 */
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
/*
 * Ссылка для мнеджера плат:
 * http://arduino.esp8266.com/stable/package_esp8266com_index.json
 */
#ifndef STASSID
#define STASSID "PROPELLER"
#define STAPSK  ""
#endif
//конфигурация IP
IPAddress local_IP(192,168,1,1);
IPAddress gateway(192,168,1,1);
IPAddress subnet(255,255,255,0);

const unsigned int localPort = 8080, //локальный порт связи
          //аппаратная ШИМ используется только для индикации LED
          R_HPWM = 15, //разрядность аппаратной ШИМ, 15-65535 / по умолчанию 1023
          //F_HPWM = 1000, //частота аппаратной ШИМ, 100 Hz-40 kHz / по умолчанию 1 kHz
          dT_PWM = 50, //"квант" времени программной ШИМ, ms
          N_PWM = 10, //кол-во "квантов" в периоде программной ШИМ
          T_MIN = 4000; //мин. длительность интервала регулирования, ms; T_MIN>>N_PWM*dT_PWM
          //T_MIN=min of <input id='rng2'> in html

const uint8_t OUT_PIN0 = 16, //D0, RUN
              OUT_PIN1 = 5, //D1, SWING
              LED_PIN = 2;  //onboard LED

unsigned long t, t_, t__, tPWM;

unsigned int pwmOut = 6, //<=R_HPWM, pwmOut=value of <input id='rng1'> in html
             pwmMax = pwmOut,
             interval = 10, //s, interval=value of <input id='rng2'> in html
             timeStep, //длительность шага регулирования
             timeOff = 360*60, //min -> s
             n, //счетчик секунд для таймера выключения
             nPWM; //счетчик программной ШИМ
             
uint8_t mod; //режим

boolean powerOn, timerOn, speedUp;

ESP8266WebServer server(localPort); //старт сервера на локальном порту

//обработчик запроса корневого URL
void handleRoot() {
 //содержимое HTML
 String content = F("<!DOCTYPE html><html><head><title>FanControl</title></head><body><script>function state(){var c1=document.getElementById('chk1');var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('chk1').innerHTML=this.responseText;}};xhttp.open('GET','set?p=st&state='+c1.checked,true);xhttp.send();if(!c1.checked) document.getElementById('chk2').checked=false;} function swing(){var c2=document.getElementById('chk2');var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('chk2').innerHTML=this.responseText;}};xhttp.open('GET','set?p=sw&swing='+c2.checked,true);xhttp.send();} function timer(){var c3=document.getElementById('chk3');var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('chk3').innerHTML=this.responseText;}};xhttp.open('GET','set?p=tm&timer='+c3.checked,true);xhttp.send();setTimeout(()=>{set_tim();},50);} function speed(){var r1=document.getElementById('rng1');var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('rng1').innerHTML=this.responseText;}};xhttp.open('GET','set?p=sp&speed='+r1.value,true);xhttp.send();var i1=document.getElementById('sind');i1.value=r1.value;} function interval(){var r2=document.getElementById('rng2');var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('rng2').innerHTML=this.responseText;}};xhttp.open('GET','set?p=int&int='+r2.value,true);xhttp.send();var i2=document.getElementById('tind');i2.value=r2.value;} function mode(){var m=document.getElementById('mod').selectedIndex;if(m==0){document.getElementById('rng2').style.visibility='hidden';document.getElementById('p1').style.visibility='hidden';}else{document.getElementById('rng2').style.visibility='visible';document.getElementById('p1').style.visibility='visible';} var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('mod').innerHTML=this.responseText;}};xhttp.open('GET','set?p=mod&mod='+m, true);xhttp.send();} function set_tim(){var t=document.getElementById('tim');var xhttp=new XMLHttpRequest();xhttp.onreadystatechange=function(){if(this.readyState==4&&this.status==200){document.getElementById('tim').innerHTML=this.responseText;}};xhttp.open('GET','set?p=tim&time='+tim.value,true);xhttp.send();}</script><h1>FanControl</h1><p><label><input id='chk1' type='checkbox' onchange='state()'>ON|OFF</label>&nbsp;<label><input id='chk2' type='checkbox' onchange='swing()'>SWING</label></p><p><label><input id='chk3' type='checkbox' onchange='timer()'>OFF Timer&nbsp;</label><label><input id='tim' maxlength=5 value='360' onkeyup='set_tim()' style='width:50px'>&nbsp;min</label></p><p>Mode&nbsp;<select id='mod' onchange='mode()'><option>Simp</option><option>Inc</option><option>Dec</option><option>Inc-Dec</option><option>Rand</option></select></p><p>Speed&nbsp;<input id='sind' maxlength=3 value='6' style='width:50px;border:0' readonly></p><input id='rng1' type='range' min='1' max='10' value='6' width='50px' onchange='speed()'><p id='p1' style='visibility:hidden'>Interval,&nbsp;s&nbsp;<input id='tind' maxlength=2 value='10' style='width:50px;border:0' readonly></p><input id='rng2' type='range' min='4' max='30' value='10' step='2' width='50px' onchange='interval()' style='visibility:hidden'></body></html>");
 server.send(200, "text/html", content); //отправка страницы
}
//обработчик запросов аргументов
void handleSet() {
   //String fb = ""; //feedback
   String parm = server.arg("p"); //тип параметра
   Serial.println(parm);

   if(parm == "st") { //символьное обозначение параметров удобно при отладке
     String state = server.arg("state"); //параметр state
     //соответствует запросу  xhttp.open("GET", "set?p=xxx&pp=yyy", true);
     Serial.println(state);
     if(state == "true") { //вкл. питание
      powerOn = true;
      //Serial.println("PowerON");
     } else { //выкл.
      powerOff(); //откл. питание
      //Serial.println("PowerOFF");  
     }
  } else if (parm == "sw") { //параметр sw (swing)
    String swing = server.arg("swing");
    Serial.println(swing);
    if(swing == "true") swingOn(true); //вкл. swing
      else swingOn(false); //откл. swing
  } else if (parm == "tm") { //параметр tm (timer)
    String timer = server.arg("timer");
    Serial.println(timer);
    if (timer == "true") timerOn = true; //таймер вкл.
      else timerOn = false; //таймер выкл.
  } else if (parm == "sp") { //параметр sp (speed)
    String speed_ = server.arg("speed");
    Serial.println(speed_);
    pwmMax = speed_.toInt(); 
    pwmOut = pwmMax; //установка скорости вращения
  } else if (parm == "int") { //параметр int (interval)
    String int_ = server.arg("int");
    Serial.println(int_);
    interval = int_.toInt();
    interval = 1000*interval;
    timeStep = round(interval/(float)R_HPWM); //длительность шага регулирования
  } else if (parm == "tim") { //параметр tim (время откл. по таймеру)
    String time_ = server.arg("time");
    Serial.println(time_);
    timeOff = 60*time_.toInt(); //min -> s
  } else if (parm == "mod") { //параметр mod (режим)
    String mode_ = server.arg("mod");
    Serial.println(mode_);
    mod = mode_.toInt();
  }
}

/** DX - GPIO
 *  TX - 1
 *  RX - 3 
 *  D8 - 15
 *  D7 - 13
 *  D6 - 12
 *  D5 - 14
 *  D4 - 2 LED
 *  D3 - 0
 *  D2 - 4
 *  D1 - 5
 *  D0 - 16
 */

void setup() {
  Serial.begin(19200);

  //PWM setup
  analogWriteRange(R_HPWM); //1023 default
  //analogWriteFreq(F_HPWM); //1 kHz default

  pwmOut = 5;

  pinMode(OUT_PIN0, OUTPUT); //software PWM
  pinMode(OUT_PIN1, OUTPUT);
  digitalWrite(OUT_PIN0, HIGH);
  digitalWrite(OUT_PIN1, HIGH);
  analogWrite(LED_PIN, R_HPWM-pwmOut);
  
  WiFi.mode(WIFI_AP); //режим точки доступа
  WiFi.softAPConfig(local_IP, gateway, subnet); //конфигурация IP
  WiFi.softAP(STASSID);

  IPAddress myIP = WiFi.softAPIP(); //IP
  Serial.print("HotSpot IP:");
  Serial.println(myIP);
  
  //назначение обработчиков запросов
  server.on("/", handleRoot);
  server.on("/set", handleSet);
  
  server.begin(); //старт сервера
  Serial.println("HTTP server started");

  interval = 1000*interval; //ms
  timeStep = round(interval/(float)R_HPWM);
}

void loop() {
  t = millis();
  server.handleClient(); 
  delay(1);
  //расчет pwmOut по выбранному режиму
  //mode: 0-usual, 1-speedUp, 2-speedDown, 3-speedUpDown, 4-random
  if (mod == 1) {
    if (t_ < t) {
      t_ = t + timeStep;
      if (pwmOut < pwmMax) pwmOut++; else pwmOut = 0;
      //Serial.println(pwmOut);
    }
  } else if (mod == 2) {
    if (t_ < t) {
      t_ = t + timeStep;
      if (pwmOut > 0) pwmOut--; else pwmOut = pwmMax;
      //Serial.println(pwmOut);
    }
  } else if (mod == 3) {
    if (t_ < t) {
      t_ = t + timeStep;
      if (pwmOut < pwmMax && speedUp) pwmOut++;
      if (pwmOut == pwmMax) speedUp = false;
      if (pwmOut > 0 && !speedUp) pwmOut--;
      if (pwmOut == 0) speedUp = true;
      //Serial.println(pwmOut);
    } 
  } else if (mod == 4) {
    if (t_ < t) {
      unsigned int t0 = random(T_MIN, interval);
      t_ = t + t0;
      pwmOut = random(0, pwmMax);
      //Serial.print("t:");
      //Serial.println(t0);
      //Serial.println(pwmOut);
    }
  }

  if (timerOn) {
    if (t__ < t) {
      t__ = t + 1000;
      n++; //кол-во секунд с момента вкл. таймера
    }   
  } else n = 0; //обнулить счетчик секунд
  //откл. по таймеру
  if (powerOn && n >= timeOff) powerOff(); //однократное выполнение условия
    
  //software PWM
  //выполняется каждые dT_PWM ms
  //при powerOn = true;
  if (powerOn && tPWM < t) {
    tPWM = t + dT_PWM;
    if (nPWM < N_PWM) { //текущее число шагов
      if (nPWM < pwmOut) digitalWrite(OUT_PIN0, LOW); //вкл.
        else digitalWrite(OUT_PIN0, HIGH); //откл.
      nPWM++;
      analogWrite(LED_PIN, R_HPWM-pwmOut); //индикация уровня 
    } else nPWM = 0; //nPWM>=N_PWM
  }
}
//функция отключения питания
void powerOff(){
  //pwmOut = 0;
    digitalWrite(OUT_PIN0, HIGH);
    digitalWrite(OUT_PIN1, HIGH);
    analogWrite(LED_PIN, R_HPWM);
    powerOn = false;
    timerOn = false;
    n = 0;
}
//функция управления swing
void swingOn(boolean swOn){
    if(powerOn && swOn) digitalWrite(OUT_PIN1, LOW);
      else digitalWrite(OUT_PIN1, HIGH);
}
